package com.interdataworking.gateway.xml;

import com.interdataworking.*;
import com.interdataworking.mm.*;
import com.interdataworking.mm.alg.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import java.util.*;
import edu.stanford.db.rdf.query.*;
import edu.stanford.db.rdf.query.memory.*;
//import edu.stanford.db.rdf.query.orderings.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;

//import edu.stanford.db.rdf.util.Aggregator;

/**
 * Given a set of forward and backward arcs, generates a map of
 * <obj, dependent obj>
 **/

public class DependenciesXSD implements UntypedGateway {

  XSDVocabulary XSD = new XSDVocabulary();
  XMLVocabulary XML = new XMLVocabulary();

  // type references
  Set ref, aggProp, elements;

  public DependenciesXSD() {

    ref = new HashSet();
    ref.add(XSD.type);
    ref.add(XSD.base);
    ref.add(XSD.ref);

    aggProp = new HashSet();
    aggProp.add(XML.tagName);
    aggProp.add(RDF.type);
    aggProp.add(XML.usesNamespace);

    elements = new HashSet();
    elements.add(XSD.element);
    elements.add(XSD.attribute);
    elements.add(XSD.complexType);
    elements.add(XSD.simpleType);
  }

  protected QueryIterator[] Q_COMPLEX_TYPE = {
    new FindIterator("E", XML.tagName, XSD.complexType),
    new FindIterator("E", XSD.name, "N")
  };

  protected QueryIterator[] Q_SIMPLE_TYPE = {
    new FindIterator("E", XML.tagName, XSD.simpleType),
    new FindIterator("E", XSD.name, "N")
  };

  // FIXME: ref is not supported properly
  protected QueryIterator[] Q_ELEMENT = {
    new FindIterator("E", XML.tagName, XSD.element),
    new OptionalIterator(new FindIterator("E", XSD.name, "N"))
  };

  protected QueryIterator[] Q_ATTRIBUTE = {
    new FindIterator("E", XML.tagName, XSD.attribute),
    new OptionalIterator(new FindIterator("E", XSD.name, "N"))
  };

  public Resource getRoot(Model model) throws ModelException {

    return RDFUtil.getSubject(model, RDF.type, XML.XMLDOM);
  }

  public Resource getParent(Model model, Resource r) throws ModelException {

    Resource p = RDFUtil.getSubject(model, XML.child, r);
    //    System.err.println("--- parent of " + r + " is " + p + " with tagName " + getTagName(model, p));
    return p;
  }

  public List getChildren(Model model, Resource r) throws ModelException {

    return RDFUtil.getObjects(model, r, XML.child);
  }

  public Resource getTagName(Model model, Resource r) throws ModelException {

    if(r == null)
      return null;
    return RDFUtil.getObjectResource(model, r, XML.tagName);
  }

  public Literal getTypeName(Model model, Resource r) throws ModelException {

    return getElementName(model, r);
  }

  public Literal getElementName(Model model, Resource r) throws ModelException {

    if(r == null)
      return null;
    return (Literal)RDFUtil.getObject(model, r, XSD.name);
  }

  public Literal getElementNameOrRef(Model model, Resource r) throws ModelException {

    if(r == null)
      return null;
    Literal l = (Literal)RDFUtil.getObject(model, r, XSD.name);
    if(l == null)
      l = (Literal)RDFUtil.getObject(model, r, XSD.ref);
    return l;
  }

  public Resource getEnclosingSchema(Model model, Resource r) throws ModelException {

    while(r != null) {

      r = getParent(model, r);
      if(XSD.schema.equals(getTagName(model, r)))
	return r;
    }
    return null;
  }

  public Resource getEnclosingNamedType(Model model, Resource r) throws ModelException {

    while(r != null) {

      r = getParent(model, r);
      Resource tagName = getTagName(model, r);
      if(XSD.complexType.equals(tagName) || XSD.simpleType.equals(tagName)) {

	if(getTypeName(model, r) != null)
	  return r;
      }
    }
    return null;
  }

  public boolean isType(Model model, Resource r) throws ModelException {

    Resource tagName = getTagName(model, r);
    if(XSD.complexType.equals(tagName) || XSD.simpleType.equals(tagName))
      return true;
    return false;
  }

  public boolean isComplexType(Model model, Resource r) throws ModelException {

    Resource tagName = getTagName(model, r);
    if(XSD.complexType.equals(tagName))
      return true;
    return false;
  }

  public boolean isSimpleType(Model model, Resource r) throws ModelException {

    Resource tagName = getTagName(model, r);
    if(XSD.simpleType.equals(tagName))
      return true;
    return false;
  }

  public boolean isNamedType(Model model, Resource r) throws ModelException {

    if(!isType(model, r))
      return false;

    return getTypeName(model, r) != null;
  }

  public boolean isSchema(Model model, Resource r) throws ModelException {

    Resource tagName = getTagName(model, r);
    if(XSD.schema.equals(tagName))
      return true;
    return false;
  }

  public Resource getEnclosingNamedElement(Model model, Resource r) throws ModelException {

    while(r != null) {

      r = getParent(model, r);
      Resource tagName = getTagName(model, r);
      if(elements.contains(tagName)) {

	if(getTypeName(model, r) != null)
	  return r;
      }
    }
    return null;
  }

  public Resource getEnclosingElement(Model model, Resource r) throws ModelException {

    while(r != null) {

      r = getParent(model, r);
      Resource tagName = getTagName(model, r);
      if(elements.contains(tagName)) {

	return r;
      }
    }
    return null;
  }

  void addDepAlongTheWayToParent(Model model, Resource child, Resource parent, Collection result) throws ModelException {

    while(child != null && !child.equals(parent)) {

      Resource r = getParent(model, child);
      MapPair pair = new MapPair(child, r);
      result.add(pair);
      child = r;
    }
  }

  public Resource getNamedElement(Model model, Literal name) throws ModelException {

    return RDFUtil.getSubject(model, XSD.name, name);
  }


  public Resource findType(Model model, Literal name) throws ModelException {

    NodeIterator it = new NodeResultSet(Q_COMPLEX_TYPE, model);
    it.set("N", name);

    if(it.next())
      return (Resource)it.get("E");

    it = new NodeResultSet(Q_SIMPLE_TYPE, model);
    it.set("N", name);

    if(it.next())
      return (Resource)it.get("E");

    return null;
  }

  /**
   * selects all information that is minimal and absolutely necessary
   * for the referenced nodes to make sense
   */
  public List getDependencies(Model model) throws ModelException {

    Set result = new HashSet();

    Set doneChildrenDep = new HashSet(); // already children of these elements

    NodeFactory f = model.getNodeFactory();

    for(Enumeration en = model.elements(); en.hasMoreElements();) {

      Statement st = (Statement)en.nextElement();
      if(ref.contains(st.predicate())) {

	Resource typeNode = getNamedElement(model, (Literal)st.object());

	//	System.err.println("Resolved type in " + st + " to " + typeNode);

	if(typeNode != null) {
	  MapPair p = new MapPair(st.subject(), typeNode);
	  result.add(p);
	  //	  addDescendantsOfDep(st.subject(), typeNode, model, result);
	}

	//	addSchemaDep(st.subject(), model, result);
      }

      if(aggProp.contains(st.predicate())) {

	MapPair p = new MapPair(st.subject(), st.object());
	result.add(p);
      }
    }

    // add type -> schema

    // COMPLEX AND SIMPLE TYPES

    {
      NodeIterator it = new NodeResultSet(Q_COMPLEX_TYPE, model);
      
      while(it.next()) {
	Resource type = (Resource)it.get("E");
	//	addSchemaDep(type, model, result);
	//	addChildrenDep(type, model, result);
      }
      
      it = new NodeResultSet(Q_SIMPLE_TYPE, model);

      while(it.next()) {
	Resource type = (Resource)it.get("E");
	//	addSchemaDep(type, model, result);
	addChildrenDep(type, model, result, doneChildrenDep); // fetch complete definition of simple types...
      }
    }

    // ELEMENTS

    {
      NodeIterator it = new NodeResultSet(Q_ELEMENT, model);

      while(it.next()) {

	//	String name = it.get("N").toString();
	Resource node = (Resource)it.get("E");

	Resource type = getEnclosingNamedType(model, node);

	// not needed since complete path to root is added anyway below in addParentDep
	//	if(type != null)
	//	  addDepAlongTheWayToParent(model, node, type, result);

	addChildrenDep(node, model, result, doneChildrenDep, 0, 1); // fetch complete definition of elements... The content of the element is essentially its type...
      }
    }
    
    // ATTRIBUTES

    // not needed since complete path to root is added anyway below in addParentDep
    /*
    {
      NodeIterator it = new NodeResultSet(Q_ATTRIBUTE, model);

      while(it.next()) {

	//	String name = it.get("N").toString();
	Resource node = (Resource)it.get("E");

	Resource type = getEnclosingNamedType(model, node);

	if(type != null)
	  addDepAlongTheWayToParent(model, node, type, result);
      }
    }
    */

    Resource root = getRoot(model);
    //    System.err.println("==== root is : " + root + " " + getTagName(model, root));

    //    addSchemaDep(isSchema(model, root) ? root : null, root, model, result);
    addParentDep(root, model, result);

    return new ArrayList(result);
  }

//   void addRequired(Model model, Collection result) throws ModelException {
//   }

  void addParentDep(Resource r, Model model, Collection result) throws ModelException {

    List children = getChildren(model, r);

    for(Iterator it = children.iterator(); it.hasNext();) {

      Resource child = (Resource)it.next();

      MapPair p = new MapPair(child, r);
      result.add(p);

      addParentDep(child, model, result);
    }
  }

  public boolean isElement(Model model, Resource r) throws ModelException {

    Resource tagName = getTagName(model, r);
    return elements.contains(tagName);
  }

  void addSchemaDep(Resource schema, Resource r, Model model, Collection result) throws ModelException {

    List children = getChildren(model, r);

    for(Iterator it = children.iterator(); it.hasNext();) {

      Resource child = (Resource)it.next();
      Resource tagName = getTagName(model, child);

      //      System.err.println("==== child " + tagName + " of schema " + schema);

      if(XSD.schema.equals(tagName))
	addSchemaDep(child, child, model, result);

      else if(schema != null && elements.contains(tagName)) { // that's a schema element

	MapPair p = new MapPair(child, schema);
	result.add(p);
      } else
	addSchemaDep(schema, child, model, result);
    }
  }

  void addChildrenDep(Resource r, Model model, Collection result, Set done) throws ModelException {

    addChildrenDep(r, model, result, done, 0, Integer.MAX_VALUE);
  }

  void addChildrenDep(Resource r, Model model, Collection result, Set done, int depth, int maxDepth) throws ModelException {
    
    if(depth++ >= maxDepth)
      return;

    if(!done.add(r))
      return; // already added this element

    List children = getChildren(model, r);

    for(Iterator it = children.iterator(); it.hasNext();) {

      Resource child = (Resource)it.next();
      MapPair p = new MapPair(r, child);
      result.add(p);
      addChildrenDep(child, model, result, done, depth, maxDepth);
    }
  }

  // adds dependencies of r to all children of parent
  void addDescendantsOfDep(Resource r, Resource parent, Model model, Collection result) throws ModelException {

    List children = getChildren(model, parent);

    for(Iterator it = children.iterator(); it.hasNext();) {

      Resource child = (Resource)it.next();
      MapPair p = new MapPair(r, child);
      result.add(p);
      addDescendantsOfDep(r, child, model, result);
    }
  }

  /*
  void addSchemaDep(Resource r, Model model, Collection result) throws ModelException {

	// make schema dependent on this one
	Resource schemaNode = getEnclosingSchema(model, r);

	//	System.err.println("Enclosing schema node for " + r + " is " + schemaNode);

	if(schemaNode != null) {
	  MapPair p = new MapPair(r, schemaNode);
	  result.add(p);
	}
  }
  */

  public List execute(List input) throws ModelException {

    Model src = (Model)input.get(0);

    Model result = MapPair.asModel(src, getDependencies(src));

    ArrayList l = new ArrayList();
    l.add(result);
    return l;
  }
  

  public int getMinInputLen() { return 1; }

  public int getMaxInputLen() { return 1; }

  public int getMinOutputLen() { return 1; }

  public int getMaxOutputLen() { return 1; }
  
}
