/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package edu.stanford.db.rdf.query.memory;

import org.w3c.rdf.model.*;
import edu.stanford.db.rdf.query.*;
import org.w3c.rdf.util.RDFUtil;
import org.w3c.rdf.util.SetOperations;

import java.util.*;

/**
 * This is a query iterator.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu> */

public class FindIterator implements QueryIterator {

  public static final int STRAIGHT = 0;
  public static final int NEGATED = 1;
  public static final int OPTIONAL = 2;

    Resource rs,rp,ri,preset_ri;
    RDFNode ro;
    String vs,vp,vo,vi;

    Model model;
    boolean unmutableModel; // model cannot be changed on reset
    Enumeration stmt;

    int cost;

  int mode = STRAIGHT;

  // if true, this iterator acts as a dummy, i.e. delivers exactly one
  // binding with unchanged variables
  boolean passThroughMode;

  public QueryIterator duplicate() {

    FindIterator it = new FindIterator();
    it.rs = rs;
    it.rp = rp;
    it.ri = ri;
    it.ro = ro;
    it.vs = vs;
    it.vp = vp;
    it.vo = vo;
    it.vi = vi;
    it.model = model;
    it.mode = mode;
    return it;
  }

  FindIterator() {
  }

    public FindIterator(Model model, Object s, Object p, Object o, Object id, int mode) {

	this(model, s, p, o, id);
	this.mode = mode;
    }

    public FindIterator(Model model, Object s, Object p, Object o, int mode) {

	this(model, s, p, o);
	this.mode = mode;
    }

    public FindIterator(Model model, Object s, Object p, Object o, Object id) {
	this(model, s, p, o);

	if(id instanceof Resource)
	    ri = (Resource)id;
	else
	    vi = (String)id;
    }

    // if model is passed in the constructor, it cannot be changed
    // using reset()
    public FindIterator(Model model, Object s, Object p, Object o) {

	this.model = model;

	if(model != null)
	    unmutableModel = true;

	if(s instanceof Resource)
	    rs = (Resource)s;
	else
	    vs = (String)s;
	
	if(p instanceof Resource)
	    rp = (Resource)p;
	else
	    vp = (String)p;

	if(o instanceof RDFNode)
	    ro = (RDFNode)o;
	else
	    vo = (String)o;
    }


    public FindIterator(Object s, Object p, Object o) {

	this(null, s, p, o);
    }

    public FindIterator(Object s, Object p, Object o, Object id) {

	this(null, s, p, o, id);
    }

    public FindIterator(Object s, Object p, Object o, int mode) {

      this(null, s, p, o, mode);
    }

    public FindIterator(Object s, Object p, Object o, Object id, int mode) {

	this(null, s, p, o, id, mode);
    }

    public int getCost() {
	
	return cost;
    }

    public Enumeration getVariables() {

	Hashtable vars = new Hashtable();
	
	if(vs != null)
	    vars.put(vs, vs);
	if(vp != null)
	    vars.put(vp, vp);
	if(vo != null)
	    vars.put(vo, vo);
	if(vi != null)
	    vars.put(vi, vi);

	return vars.elements();
    }

    public void reset(Hashtable vars, Model m) throws ModelException {

	//	System.err.println("Set model of size: " + m.size());

	if(!unmutableModel && m != null)
	    model = m;

	// we cannot "find" for IDs directly, so remember the preset id
	preset_ri = null;

	if(model == null)
	  throw new ModelException("FindIterator: model not set!");

	// issue query
	try {
	    Model list = model.find(vs != null ? (Resource)vars.get(vs) : rs,
				    vp != null ? (Resource)vars.get(vp) : rp,
				    vo != null ? (RDFNode)vars.get(vo) : ro);
	    cost++;

	    switch(mode) {

	    case NEGATED:

	      if(list.size() == 0)
		// allow exacty one binding
		passThroughMode = true;
	      // else: this binding attempt failed!
	      stmt = null;
	      break;

	    case OPTIONAL:

	      if(list.size() == 0) {
		// nevertheless allow exactly one binding
		passThroughMode = true;
		stmt = null;
	      } else
		stmt = list.elements();
	      break;

	    default: // STRAIGHT
	      stmt = list.elements();
	    }

	    if(vi != null)
	      preset_ri = (Resource)vars.get(vi);

// 	    println("Found " + list.size() + " triples for " +
// 		    (vs != null ? (Resource)vars.get(vs) : rs) + "," +
// 		    (vp != null ? (Resource)vars.get(vp) : rp) + "," +
// 		    (vo != null ? (RDFNode)vars.get(vo) : ro));
	} catch (ClassCastException cce) {
	    // in case literal is used as subject
	    stmt = null;
	}
    }

    private void println(String msg) {

      System.err.println(toString() + ": " + msg);
    }

    public boolean next(Hashtable vars) throws ModelException {

      if(passThroughMode) {
	passThroughMode = false;
	return true;
      }

	while(stmt != null && stmt.hasMoreElements()) {

	    cost++;
	    Statement t = (Statement)stmt.nextElement();

	    if((preset_ri != null && !t.equals(preset_ri)) ||
	       (ri != null && !t.equals(ri)))
		continue;

	    if(vi != null) {
		vars.put(vi, t);
		//		println("Put: " + vi + "=" + t);
	    }

	    if(vs != null) {
		vars.put(vs, t.subject());
		//		println("Put: " + vs + "=" + t.subject());
	    }
	    if(vp != null) {
		vars.put(vp, t.predicate());
		//		println("Put: " + vp + "=" + t.predicate());
	    }
	    if(vo != null) {
		vars.put(vo, t.object());
		//		println("Put: " + vo + "=" + t.object());
	    }
	    return true;
	}
	return false;
    }

}
